from lark import Lark, Transformer
from mel_ast import *


parser = Lark('''
    %import common.NUMBER
    %import common.CNAME
    %import common.NEWLINE
    %import common.WS

    %ignore WS

    COMMENT: "/*" /(.|\\n|\\r)+/ "*/"
        |  "//" /(.)+/ NEWLINE
    %ignore COMMENT

    num: NUMBER
    ident: CNAME

    ?group: num
        | ident
        | "(" expr ")"

    ?mult: group
        | mult "*" group    -> mul
        | mult "/" group    -> div

    ?add: mult
        | add "+" mult      -> add
        | add "-" mult      -> sub
        
    ?expr: add

    ?stmt: "input" ident    -> input
        | "output" expr      -> output
        | ident "=" expr     -> assign

    ?prog: stmt*            -> stmt_list

    ?start: prog
''', start="start")


class MelASTBuilder(Transformer):
    def _call_userfunc(self, tree, new_children=None):
        # Assumes tree is already transformed
        children = new_children if new_children is not None else tree.children
        try:
            f = getattr(self, tree.data)
        except AttributeError:
            return self.__default__(tree.data, children, tree.meta)
        else:
            return f(*children)

    def __getattr__(self, item):
        if isinstance(item, str) and item.upper() == item:
            return lambda x: x

        if item in ('mul', 'div', 'add', 'sub'):
            def get_bin_op_node(*args):
                op = BinOp[item.upper()]
                return BinOpNode(op, *args)
            return get_bin_op_node
        else:
            def get_node(*args):
                cls = eval(''.join(x.capitalize() or '_' for x in item.split('_')) + 'Node')
                return cls(*args)
            return get_node


def parse(prog: str)->StmtListNode:
    prog = parser.parse(str(prog))
    prog = MelASTBuilder().transform(prog)
    return prog
